//34567890123456789012345678901234567890123456789012345678
import greenfoot.*;
import java.awt.Font;
import java.awt.Color;

/**
 * This is a modest update to the earlier TextBox class
 * named TextBoxA.
 * 
 * Any object that instantiates an object of this class
 * named TextBoxB must implement the TextBoxListener 
 * interface and define the callback method. If you don't 
 * want to do that, use the earlier version of the class
 * named TextBoxA
 * 
 * Note that an object of this class has no concept of
 * focus.  Therefore, it must not be used concurrently
 * with another object that is competing for keystrokes.
 * 
 * This Actor class can be used to instantiate a text box
 * that will accept text input from the user and return
 * that text as a String object when the getText method is
 * called on the object. This updated version adds the
 * ability to register the user object as a listener and
 * to notify the listener when text is ready for 
 * retrieval, thus eliminating the polling requirement of
 * the  earlier version. This version can be used either 
 * way by either:
 * 
 * 1 - Calling the getText method in the act method and 
 * leaving the callback method empty.
 * 
 * 2 - Calling the getText method from within the callback
 * method and not calling the getText method in the act\
 * method.
 * 
 * The getText method returns null if the user hasn't yet
 * stroked the Enter key.  When the user strokes the
 * Enter key, the callback method is called to notify the
 * user object that text is ready to be retrieved. 
 * 
 * The string contents of the text box will be returned 
 * the next time the getText method is called. When the
 * getText method is called, the text box is cleared and 
 * made ready to receive new text data.
 * 
 * The user object should either poll the text box by
 * calling the getText method inside the act method, or
 * should call the getText method when its callback method
 * is called, but not both.
 * 
 * The text box supports the use of the backspace key to
 * delete text that has been entered one character at a 
 * time from the right end of the text string. It does not
 * support arrow keys or functions keys. If those keys are
 * stroked, the resulting string will contain the names
 * of those keys.
 * 
 * @author Dick Baldwin 
 * @version 07/18/08
 */
public class TextBoxB extends Actor{
  private Labels textLabel;//text box image
  private StringBuffer string = new StringBuffer();
  private String output = null;//output saved here
  private int halfWidth;//half width of initial label.
  private int halfHeight;//half height of initial label
  private String key;//used to store a keystroke
  private int leftEdge;//left edge of text box
  private int vertCenter;//vertical center of text box
  private boolean dontAskAgain = false;//see the code
  private int points;//point size for the text
  
  //A reference back to the object that instantiated this
  // TextBoxB object.
  TextBoxListener listener;
  //----------------------------------------------------//
  
  public TextBoxB(String initialText,
                  int points,
                  TextBoxListener listener){
    this.listener = listener;//save ref to listener
    if(initialText.equals("")){
      initialText = " ";
    }//end if
    this.points = points;
    textLabel = makeLabel(initialText);
    halfWidth = textLabel.getWidth()/2;
    halfHeight = textLabel.getHeight()/2;
    //This TextBoxB object is invisible but the label that
    // represents it is not invisible.
    setImage(new GreenfootImage(1,1));
  }//end constructor
  //----------------------------------------------------//
  
  public void act(){
    
    //Get and save the x-coordinate of the left edge and
    // the y-coordinate of the center of the label at the
    // first opportunity after it is added to the world.
    // Only get and save those values once.
    if((textLabel.getWorld() != null) && 
                                 (dontAskAgain == false)){
      leftEdge = textLabel.getX() - halfWidth;
      vertCenter = textLabel.getY();
      dontAskAgain = true;
    }//end if
    
    //Now process keystrokes.
    key = Greenfoot.getKey();//get most recent keystroke
    if(key != null){
      //Most of the time key will be null unless the user
      // types extremely fast.
      if(key.equals("enter")){
        //Signal that output data is ready to be used by
        // replacing null in the output with an actual
        // String.
        output = string.toString();
        //Also notify the listener that the text is ready.
        listener.callback(this);
      }else{
        if(key.equals("backspace")){
          //Delete character on the end of the string, but
          // only if there are characters in the 
          // StringBuffer object.
          if(string.length() > 0){
            string.deleteCharAt(string.length() - 1);
          }//end if
        }else{
          if(key.equals("space")){
            key = " ";
          }//end if
          //Append this keystroke to the string that is
          // being constructed. Note that objects of the
          // TextBoxB class aren't designed to deal with
          // arrow keys, function keys, etc. The names of
          // those keys will simply be appended to the
          // string that is being constructed if those
          // keys are stroked.
          string.append(key);
        }//end else
        
        //Update the label to reflect the current state of
        // the string that is being constructed.
        //First remove the current version of the label
        // from the world so that it can be replaced by a
        // new version. Then create the new version and 
        // add it to the world.
        getWorld().removeObject(textLabel);
        textLabel = makeLabel(string.toString());
        //The following code causes the left edge of the
        // label to stay in the same location and causes
        // the label to extend to the right as its width
        // increases due to the addition of more
        // characters. Note that when getWidth is called
        // here, it returns the width of the current 
        // version of the label as opposed to the value
        // of halfWidth that is based on the initial
        // label.
        getWorld().addObject(textLabel,
            leftEdge + textLabel.getWidth()/2,vertCenter);
      }//end else
    }//end if keystroke is null
  }//end act
  //----------------------------------------------------//
  
  //The purpose of this method is to construct a new
  // label for an incoming string.
  Labels makeLabel(String text){
    return new Labels(
      text,//Text to be displayed in the label
      "SansSerif",//Can be specific font or family
      Font.BOLD,//BOLD, ITALIC, PLAIN, etc
      points,//Text size in points saved as instance var.
      Color.BLACK,//Text color
      null,//Picture for left and/or right
      true,//Left or right justify - N/A
      true,//Pictures on both sides of text - N/A
      true,//Solid outline
      Color.BLACK,//Solid outline color
      null);
  }//end makeLabel
  //----------------------------------------------------//
  
  //The purpose of this method is to return the string
  // that has been constructed. Note that this method
  // returns null until the user strokes the Enter key,
  // following which it returns a reference to a String
  // object. Also note that the data will remain in the
  // output variable until this method is called, at
  // which time the code will prepare the TextBoxB object
  // for data entry and construction of a new string and
  // return the old one.
  public String getText(){
    if(output == null){
      return output;
    }else{
      //Save the current string.
      String temp = output.toString();
      //Prepare to construct a new string.
      string = new StringBuffer();
      output = null;
      getWorld().removeObject(textLabel);
      textLabel = makeLabel(" ");
      getWorld().addObject(
               textLabel,leftEdge + halfWidth,vertCenter);
      return temp;
    }//end else
  }//end getText()
  //----------------------------------------------------//
  
  //This method returns the current label. The primary
  // use of this method is to make it possible for an
  // Actor to add the label to the world after the Actor
  // has been constructed.
  Labels getTextBoxLabel(){
    return textLabel;
  }//end getTextBoxLabel
  //----------------------------------------------------//

}//end class TextBoxB
